from databricks.labs.blueprint.logger import install_logger
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.sdk.core import with_user_agent_extra

install_logger()
with_user_agent_extra("cmd", "install")

if __name__ == "__main__":
    logger = get_logger(__file__)
    logger.setLevel("INFO")

    logger.info("Successfully Setup Remorph Components Locally")



import logging
import sys
import os
from datetime import datetime
from uuid import uuid4

from pyspark.errors import PySparkException
from pyspark.sql import DataFrame, SparkSession
from sqlglot import Dialect

from databricks.labs.lakebridge.config import (
    DatabaseConfig,
    TableRecon,
    ReconcileConfig,
    ReconcileMetadataConfig,
)
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_dialect
from databricks.labs.lakebridge.reconcile.compare import (
    capture_mismatch_data_and_columns,
    reconcile_data,
    join_aggregate_data,
    reconcile_agg_data_per_rule,
)
from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.connectors.source_adapter import create_adapter
from databricks.labs.lakebridge.reconcile.exception import (
    DataSourceRuntimeException,
    InvalidInputException,
    ReconciliationException,
)
from databricks.labs.lakebridge.reconcile.query_builder.aggregate_query import AggregateQueryBuilder
from databricks.labs.lakebridge.reconcile.query_builder.count_query import CountQueryBuilder
from databricks.labs.lakebridge.reconcile.query_builder.hash_query import HashQueryBuilder
from databricks.labs.lakebridge.reconcile.query_builder.sampling_query import (
    SamplingQueryBuilder,
)
from databricks.labs.lakebridge.reconcile.query_builder.threshold_query import (
    ThresholdQueryBuilder,
)
from databricks.labs.lakebridge.reconcile.recon_capture import (
    ReconCapture,
    generate_final_reconcile_output,
    ReconIntermediatePersist,
    generate_final_reconcile_aggregate_output,
)
from databricks.labs.lakebridge.reconcile.recon_config import (
    Schema,
    Table,
    AggregateQueryRules,
    SamplingOptions,
    RECONCILE_OPERATION_NAME,
    AGG_RECONCILE_OPERATION_NAME,
)
from databricks.labs.lakebridge.reconcile.recon_output_config import (
    DataReconcileOutput,
    ReconcileOutput,
    ReconcileProcessDuration,
    SchemaReconcileOutput,
    ThresholdOutput,
    ReconcileRecordCount,
    AggregateQueryOutput,
)
from databricks.labs.lakebridge.reconcile.sampler import SamplerFactory
from databricks.labs.lakebridge.reconcile.schema_compare import SchemaCompare
from databricks.labs.lakebridge.transpiler.execute import verify_workspace_client
from databricks.sdk import WorkspaceClient
from databricks.labs.blueprint.installation import Installation
from databricks.connect import DatabricksSession

logger = logging.getLogger(__name__)
_SAMPLE_ROWS = 50


def validate_input(input_value: str, list_of_value: set, message: str):
    if input_value not in list_of_value:
        error_message = f"{message} --> {input_value} is not one of {list_of_value}"
        logger.error(error_message)
        raise InvalidInputException(error_message)


def main(*argv) -> None:
    logger.debug(f"Arguments received: {argv}")

    assert len(sys.argv) == 2, f"Invalid number of arguments: {len(sys.argv)}," f" Operation name must be specified."
    operation_name = sys.argv[1]

    assert operation_name in {
        RECONCILE_OPERATION_NAME,
        AGG_RECONCILE_OPERATION_NAME,
    }, f"Invalid option: {operation_name}"

    w = WorkspaceClient()

    installation = Installation.assume_user_home(w, "lakebridge")

    reconcile_config = installation.load(ReconcileConfig)

    catalog_or_schema = (
        reconcile_config.database_config.source_catalog
        if reconcile_config.database_config.source_catalog
        else reconcile_config.database_config.source_schema
    )
    filename = f"recon_config_{reconcile_config.data_source}_{catalog_or_schema}_{reconcile_config.report_type}.json"

    logger.info(f"Loading {filename} from Databricks Workspace...")

    table_recon = installation.load(type_ref=TableRecon, filename=filename)

    if operation_name == AGG_RECONCILE_OPERATION_NAME:
        return _trigger_reconcile_aggregates(w, table_recon, reconcile_config)

    return _trigger_recon(w, table_recon, reconcile_config)


def _trigger_recon(
    w: WorkspaceClient,
    table_recon: TableRecon,
    reconcile_config: ReconcileConfig,
):
    try:
        recon_output = recon(
            ws=w,
            spark=DatabricksSession.builder.getOrCreate(),
            table_recon=table_recon,
            reconcile_config=reconcile_config,
        )
        logger.info(f"recon_output: {recon_output}")
        logger.info(f"recon_id: {recon_output.recon_id}")
    except ReconciliationException as e:
        logger.error(f"Error while running recon: {e.reconcile_output}")
        raise e


def _trigger_reconcile_aggregates(
    ws: WorkspaceClient,
    table_recon: TableRecon,
    reconcile_config: ReconcileConfig,
):
    """
    Triggers the reconciliation process for aggregated data  between source and target tables.
    Supported Aggregate functions: MIN, MAX, COUNT, SUM, AVG, MEAN, MODE, PERCENTILE, STDDEV, VARIANCE, MEDIAN

    This function attempts to reconcile aggregate data based on the configurations provided. It logs the outcome
    of the reconciliation process, including any errors encountered during execution.

    Parameters:
    - ws (WorkspaceClient): The workspace client used to interact with Databricks workspaces.
    - table_recon (TableRecon): Configuration for the table reconciliation process, including source and target details.
    - reconcile_config (ReconcileConfig): General configuration for the reconciliation process,
                                                                    including database and table settings.

    Raises:
    - ReconciliationException: If an error occurs during the reconciliation process, it is caught and re-raised
      after logging the error details.
    """
    try:
        recon_output = reconcile_aggregates(
            ws=ws,
            spark=DatabricksSession.builder.getOrCreate(),
            table_recon=table_recon,
            reconcile_config=reconcile_config,
        )
        logger.info(f"recon_output: {recon_output}")
        logger.info(f"recon_id: {recon_output.recon_id}")
    except ReconciliationException as e:
        logger.error(f"Error while running aggregate reconcile: {str(e)}")
        raise e


def recon(
    ws: WorkspaceClient,
    spark: SparkSession,
    table_recon: TableRecon,
    reconcile_config: ReconcileConfig,
    local_test_run: bool = False,
) -> ReconcileOutput:
    """[EXPERIMENTAL] Reconcile the data between the source and target tables."""
    # verify the workspace client and add proper product and version details
    # TODO For now we are utilising the
    #  verify_workspace_client from transpile/execute.py file. Later verify_workspace_client function has to be
    #  refactored

    ws_client: WorkspaceClient = verify_workspace_client(ws)

    # validate the report type
    report_type = reconcile_config.report_type.lower()
    logger.info(f"report_type: {report_type}, data_source: {reconcile_config.data_source} ")
    validate_input(report_type, {"schema", "data", "row", "all"}, "Invalid report type")

    source, target = initialise_data_source(
        engine=get_dialect(reconcile_config.data_source),
        spark=spark,
        ws=ws_client,
        secret_scope=reconcile_config.secret_scope,
    )

    recon_id = str(uuid4())
    # initialise the Reconciliation
    reconciler = Reconciliation(
        source,
        target,
        reconcile_config.database_config,
        report_type,
        SchemaCompare(spark=spark),
        get_dialect(reconcile_config.data_source),
        spark,
        metadata_config=reconcile_config.metadata_config,
    )

    # initialise the recon capture class
    recon_capture = ReconCapture(
        database_config=reconcile_config.database_config,
        recon_id=recon_id,
        report_type=report_type,
        source_dialect=get_dialect(reconcile_config.data_source),
        ws=ws_client,
        spark=spark,
        metadata_config=reconcile_config.metadata_config,
        local_test_run=local_test_run,
    )

    for table_conf in table_recon.tables:
        recon_process_duration = ReconcileProcessDuration(start_ts=str(datetime.now()), end_ts=None)
        schema_reconcile_output = SchemaReconcileOutput(is_valid=True)
        data_reconcile_output = DataReconcileOutput()
        try:
            src_schema, tgt_schema = _get_schema(
                source=source, target=target, table_conf=table_conf, database_config=reconcile_config.database_config
            )
        except DataSourceRuntimeException as e:
            schema_reconcile_output = SchemaReconcileOutput(is_valid=False, exception=str(e))
        else:
            if report_type in {"schema", "all"}:
                schema_reconcile_output = _run_reconcile_schema(
                    reconciler=reconciler, table_conf=table_conf, src_schema=src_schema, tgt_schema=tgt_schema
                )
                logger.warning("Schema comparison is completed.")

            if report_type in {"data", "row", "all"}:
                data_reconcile_output = _run_reconcile_data(
                    reconciler=reconciler, table_conf=table_conf, src_schema=src_schema, tgt_schema=tgt_schema
                )
                logger.warning(f"Reconciliation for '{report_type}' report completed.")

        recon_process_duration.end_ts = str(datetime.now())
        # Persist the data to the delta tables
        recon_capture.start(
            data_reconcile_output=data_reconcile_output,
            schema_reconcile_output=schema_reconcile_output,
            table_conf=table_conf,
            recon_process_duration=recon_process_duration,
            record_count=reconciler.get_record_count(table_conf, report_type),
        )
        if report_type != "schema":
            ReconIntermediatePersist(
                spark=spark, path=generate_volume_path(table_conf, reconcile_config.metadata_config)
            ).clean_unmatched_df_from_volume()

    return _verify_successful_reconciliation(
        generate_final_reconcile_output(
            recon_id=recon_id,
            spark=spark,
            metadata_config=reconcile_config.metadata_config,
            local_test_run=local_test_run,
        )
    )


def _verify_successful_reconciliation(
    reconcile_output: ReconcileOutput, operation_name: str = "reconcile"
) -> ReconcileOutput:
    for table_output in reconcile_output.results:
        if table_output.exception_message or (
            table_output.status.column is False
            or table_output.status.row is False
            or table_output.status.schema is False
            or table_output.status.aggregate is False
        ):
            raise ReconciliationException(
                f" Reconciliation failed for one or more tables. Please check the recon metrics for more details."
                f" **{operation_name}** failed.",
                reconcile_output=reconcile_output,
            )

    logger.info("Reconciliation completed successfully.")
    return reconcile_output


def generate_volume_path(table_conf: Table, metadata_config: ReconcileMetadataConfig):
    catalog = metadata_config.catalog
    schema = metadata_config.schema
    return f"/Volumes/{catalog}/{schema}/{metadata_config.volume}/{table_conf.source_name}_{table_conf.target_name}/"


def initialise_data_source(
    ws: WorkspaceClient,
    spark: SparkSession,
    engine: Dialect,
    secret_scope: str,
):
    source = create_adapter(engine=engine, spark=spark, ws=ws, secret_scope=secret_scope)
    target = create_adapter(engine=get_dialect("databricks"), spark=spark, ws=ws, secret_scope=secret_scope)

    return source, target


def _get_missing_data(
    reader: DataSource,
    sampler: SamplingQueryBuilder,
    missing_df: DataFrame,
    catalog: str,
    schema: str,
    table_name: str,
) -> DataFrame:
    sample_query = sampler.build_query(missing_df)
    return reader.read_data(
        catalog=catalog,
        schema=schema,
        table=table_name,
        query=sample_query,
        options=None,
    )


def reconcile_aggregates(
    ws: WorkspaceClient,
    spark: SparkSession,
    table_recon: TableRecon,
    reconcile_config: ReconcileConfig,
    local_test_run: bool = False,
):
    """[EXPERIMENTAL] Reconcile the aggregated data between the source and target tables.
    for e.g., COUNT, SUM, AVG of columns between source and target with or without any specific key/group by columns
    Supported Aggregate functions: MIN, MAX, COUNT, SUM, AVG, MEAN, MODE, PERCENTILE, STDDEV, VARIANCE, MEDIAN
    """
    # verify the workspace client and add proper product and version details
    # TODO For now we are utilising the
    #  verify_workspace_client from transpile/execute.py file. Later verify_workspace_client function has to be
    #  refactored

    ws_client: WorkspaceClient = verify_workspace_client(ws)

    report_type = ""
    if report_type:
        logger.info(f"report_type: {report_type}")
    logger.info(f"data_source: {reconcile_config.data_source}")

    # Read the reconcile_config and initialise the source and target data sources. Target is always Databricks
    source, target = initialise_data_source(
        engine=get_dialect(reconcile_config.data_source),
        spark=spark,
        ws=ws_client,
        secret_scope=reconcile_config.secret_scope,
    )

    # Generate Unique recon_id for every run
    recon_id = str(uuid4())

    # initialise the Reconciliation
    reconciler = Reconciliation(
        source,
        target,
        reconcile_config.database_config,
        report_type,
        SchemaCompare(spark=spark),
        get_dialect(reconcile_config.data_source),
        spark,
        metadata_config=reconcile_config.metadata_config,
    )

    # initialise the recon capture class
    recon_capture = ReconCapture(
        database_config=reconcile_config.database_config,
        recon_id=recon_id,
        report_type=report_type,
        source_dialect=get_dialect(reconcile_config.data_source),
        ws=ws_client,
        spark=spark,
        metadata_config=reconcile_config.metadata_config,
        local_test_run=local_test_run,
    )

    # Get the Aggregated Reconciliation Output for each table
    for table_conf in table_recon.tables:
        recon_process_duration = ReconcileProcessDuration(start_ts=str(datetime.now()), end_ts=None)
        try:
            src_schema, tgt_schema = _get_schema(
                source=source,
                target=target,
                table_conf=table_conf,
                database_config=reconcile_config.database_config,
            )
        except DataSourceRuntimeException as e:
            raise ReconciliationException(message=str(e)) from e

        assert table_conf.aggregates, "Aggregates must be defined for Aggregates Reconciliation"

        table_reconcile_agg_output_list: list[AggregateQueryOutput] = _run_reconcile_aggregates(
            reconciler=reconciler,
            table_conf=table_conf,
            src_schema=src_schema,
            tgt_schema=tgt_schema,
        )

        recon_process_duration.end_ts = str(datetime.now())

        # Persist the data to the delta tables
        recon_capture.store_aggregates_metrics(
            reconcile_agg_output_list=table_reconcile_agg_output_list,
            table_conf=table_conf,
            recon_process_duration=recon_process_duration,
        )

        (
            ReconIntermediatePersist(
                spark=spark,
                path=generate_volume_path(table_conf, reconcile_config.metadata_config),
            ).clean_unmatched_df_from_volume()
        )

    return _verify_successful_reconciliation(
        generate_final_reconcile_aggregate_output(
            recon_id=recon_id,
            spark=spark,
            metadata_config=reconcile_config.metadata_config,
            local_test_run=local_test_run,
        ),
        operation_name=AGG_RECONCILE_OPERATION_NAME,
    )


class Reconciliation:

    def __init__(
        self,
        source: DataSource,
        target: DataSource,
        database_config: DatabaseConfig,
        report_type: str,
        schema_comparator: SchemaCompare,
        source_engine: Dialect,
        spark: SparkSession,
        metadata_config: ReconcileMetadataConfig,
    ):
        self._source = source
        self._target = target
        self._report_type = report_type
        self._database_config = database_config
        self._schema_comparator = schema_comparator
        self._target_engine = get_dialect("databricks")
        self._source_engine = source_engine
        self._spark = spark
        self._metadata_config = metadata_config

    def reconcile_data(
        self,
        table_conf: Table,
        src_schema: list[Schema],
        tgt_schema: list[Schema],
    ) -> DataReconcileOutput:
        data_reconcile_output = self._get_reconcile_output(table_conf, src_schema, tgt_schema)
        reconcile_output = data_reconcile_output
        if self._report_type in {"data", "all"}:
            reconcile_output = self._get_sample_data(table_conf, data_reconcile_output, src_schema, tgt_schema)
            if table_conf.get_threshold_columns("source"):
                reconcile_output.threshold_output = self._reconcile_threshold_data(table_conf, src_schema, tgt_schema)

        if self._report_type == "row" and table_conf.get_threshold_columns("source"):
            logger.warning("Threshold comparison is ignored for 'row' report type")

        return reconcile_output

    def reconcile_schema(
        self,
        src_schema: list[Schema],
        tgt_schema: list[Schema],
        table_conf: Table,
    ):
        return self._schema_comparator.compare(src_schema, tgt_schema, self._source_engine, table_conf)

    def reconcile_aggregates(
        self,
        table_conf: Table,
        src_schema: list[Schema],
        tgt_schema: list[Schema],
    ) -> list[AggregateQueryOutput]:
        return self._get_reconcile_aggregate_output(table_conf, src_schema, tgt_schema)

    def _get_reconcile_output(
        self,
        table_conf,
        src_schema,
        tgt_schema,
    ):
        src_hash_query = HashQueryBuilder(table_conf, src_schema, "source", self._source_engine).build_query(
            report_type=self._report_type
        )
        tgt_hash_query = HashQueryBuilder(table_conf, tgt_schema, "target", self._source_engine).build_query(
            report_type=self._report_type
        )
        src_data = self._source.read_data(
            catalog=self._database_config.source_catalog,
            schema=self._database_config.source_schema,
            table=table_conf.source_name,
            query=src_hash_query,
            options=table_conf.jdbc_reader_options,
        )
        tgt_data = self._target.read_data(
            catalog=self._database_config.target_catalog,
            schema=self._database_config.target_schema,
            table=table_conf.target_name,
            query=tgt_hash_query,
            options=table_conf.jdbc_reader_options,
        )

        volume_path = generate_volume_path(table_conf, self._metadata_config)
        return reconcile_data(
            source=src_data,
            target=tgt_data,
            key_columns=table_conf.join_columns,
            report_type=self._report_type,
            spark=self._spark,
            path=volume_path,
        )

    def _get_reconcile_aggregate_output(
        self,
        table_conf,
        src_schema,
        tgt_schema,
    ):
        """
        Creates a single Query, for the aggregates having the same group by columns. (Ex: 1)
        If there are no group by columns, all the aggregates are clubbed together in a single query. (Ex: 2)
        Examples:
            1.  {
                      "type": "MIN",
                      "agg_cols": ["COL1"],
                      "group_by_cols": ["COL4"]
                    },
                    {
                      "type": "MAX",
                      "agg_cols": ["COL2"],
                      "group_by_cols": ["COL9"]
                    },
                    {
                      "type": "COUNT",
                      "agg_cols": ["COL2"],
                      "group_by_cols": ["COL9"]
                    },
                    {
                      "type": "AVG",
                      "agg_cols": ["COL3"],
                      "group_by_cols": ["COL4"]
                    },
              Query 1: SELECT MIN(COL1), AVG(COL3) FROM :table GROUP BY COL4
              Rules: ID  | Aggregate Type | Column | Group By Column
                         #1,   MIN,                      COL1,     COL4
                         #2,   AVG,                     COL3,      COL4
              -------------------------------------------------------
              Query 2: SELECT MAX(COL2), COUNT(COL2) FROM :table GROUP BY COL9
              Rules: ID  | Aggregate Type | Column | Group By Column
                         #1,   MAX,                      COL2,     COL9
                         #2,   COUNT,                COL2,      COL9
          2.  {
              "type": "MAX",
              "agg_cols": ["COL1"]
            },
            {
              "type": "SUM",
              "agg_cols": ["COL2"]
            },
            {
              "type": "MAX",
              "agg_cols": ["COL3"]
            }
          Query: SELECT MAX(COL1), SUM(COL2), MAX(COL3) FROM :table
          Rules: ID  | Aggregate Type | Column | Group By Column
                     #1, MAX, COL1,
                     #2, SUM, COL2,
                     #3, MAX, COL3,
        """

        src_query_builder = AggregateQueryBuilder(
            table_conf,
            src_schema,
            "source",
            self._source_engine,
        )

        # build Aggregate queries for source,
        src_agg_queries: list[AggregateQueryRules] = src_query_builder.build_queries()

        # There could be one or more queries per table based on the group by columns

        # build Aggregate queries for target(Databricks),
        tgt_agg_queries: list[AggregateQueryRules] = AggregateQueryBuilder(
            table_conf,
            tgt_schema,
            "target",
            self._target_engine,
        ).build_queries()

        volume_path = generate_volume_path(table_conf, self._metadata_config)

        table_agg_output: list[AggregateQueryOutput] = []

        # Iterate over the grouped aggregates and reconcile the data
        # Zip all the keys, read the source, target data for each Aggregate query
        # and reconcile on the aggregate data
        # For e.g., (source_query_GRP1, target_query_GRP1), (source_query_GRP2, target_query_GRP2)
        for src_query_with_rules, tgt_query_with_rules in zip(src_agg_queries, tgt_agg_queries):
            # For each Aggregate query, read the Source and Target Data and add a hash column

            rules_reconcile_output: list[AggregateQueryOutput] = []
            src_data = None
            tgt_data = None
            joined_df = None
            data_source_exception = None
            try:
                src_data = self._source.read_data(
                    catalog=self._database_config.source_catalog,
                    schema=self._database_config.source_schema,
                    table=table_conf.source_name,
                    query=src_query_with_rules.query,
                    options=table_conf.jdbc_reader_options,
                )
                tgt_data = self._target.read_data(
                    catalog=self._database_config.target_catalog,
                    schema=self._database_config.target_schema,
                    table=table_conf.target_name,
                    query=tgt_query_with_rules.query,
                    options=table_conf.jdbc_reader_options,
                )
                # Join the Source and Target Aggregated data
                joined_df = join_aggregate_data(
                    source=src_data,
                    target=tgt_data,
                    key_columns=src_query_with_rules.group_by_columns,
                    spark=self._spark,
                    path=f"{volume_path}{src_query_with_rules.group_by_columns_as_str}",
                )
            except DataSourceRuntimeException as e:
                data_source_exception = e

            # For each Aggregated Query, reconcile the data based on the rule
            for rule in src_query_with_rules.rules:
                if data_source_exception:
                    rule_reconcile_output = DataReconcileOutput(exception=str(data_source_exception))
                else:
                    rule_reconcile_output = reconcile_agg_data_per_rule(
                        joined_df, src_data.columns, tgt_data.columns, rule
                    )
                rules_reconcile_output.append(AggregateQueryOutput(rule=rule, reconcile_output=rule_reconcile_output))

            # For each table, there could be many Aggregated queries.
            # Collect the list of Rule Reconcile output per each Aggregate query and append it to the list
            table_agg_output.extend(rules_reconcile_output)
        return table_agg_output

    def _get_sample_data(
        self,
        table_conf,
        reconcile_output,
        src_schema,
        tgt_schema,
    ):
        mismatch = None
        missing_in_src = None
        missing_in_tgt = None

        if (
            reconcile_output.mismatch_count > 0
            or reconcile_output.missing_in_src_count > 0
            or reconcile_output.missing_in_tgt_count > 0
        ):
            src_sampler = SamplingQueryBuilder(table_conf, src_schema, "source", self._source_engine)
            tgt_sampler = SamplingQueryBuilder(table_conf, tgt_schema, "target", self._target_engine)
            if reconcile_output.mismatch_count > 0:
                mismatch = self._get_mismatch_data(
                    src_sampler,
                    tgt_sampler,
                    reconcile_output.mismatch_count,
                    reconcile_output.mismatch.mismatch_df,
                    table_conf.join_columns,
                    table_conf.source_name,
                    table_conf.target_name,
                    table_conf.sampling_options,
                )

            if reconcile_output.missing_in_src_count > 0:
                missing_in_src = _get_missing_data(
                    self._target,
                    tgt_sampler,
                    reconcile_output.missing_in_src,
                    self._database_config.target_catalog,
                    self._database_config.target_schema,
                    table_conf.target_name,
                )

            if reconcile_output.missing_in_tgt_count > 0:
                missing_in_tgt = _get_missing_data(
                    self._source,
                    src_sampler,
                    reconcile_output.missing_in_tgt,
                    self._database_config.source_catalog,
                    self._database_config.source_schema,
                    table_conf.source_name,
                )

        return DataReconcileOutput(
            mismatch=mismatch,
            mismatch_count=reconcile_output.mismatch_count,
            missing_in_src_count=reconcile_output.missing_in_src_count,
            missing_in_tgt_count=reconcile_output.missing_in_tgt_count,
            missing_in_src=missing_in_src,
            missing_in_tgt=missing_in_tgt,
        )

    def _get_mismatch_data(
        self,
        src_sampler,
        tgt_sampler,
        mismatch_count,
        mismatch,
        key_columns,
        src_table: str,
        tgt_table: str,
        sampling_options: SamplingOptions,
    ):

        tgt_sampling_query = tgt_sampler.build_query_with_alias()

        sampling_model_target = self._target.read_data(
            catalog=self._database_config.target_catalog,
            schema=self._database_config.target_schema,
            table=tgt_table,
            query=tgt_sampling_query,
            options=None,
        )

        # Uses pre-calculated `mismatch_count` from `reconcile_output.mismatch_count` to avoid from recomputing `mismatch` for RandomSampler.
        mismatch_sampler = SamplerFactory.get_sampler(sampling_options)
        df = mismatch_sampler.sample(mismatch, mismatch_count, key_columns, sampling_model_target).cache()

        src_mismatch_sample_query = src_sampler.build_query(df)
        tgt_mismatch_sample_query = tgt_sampler.build_query(df)

        src_data = self._source.read_data(
            catalog=self._database_config.source_catalog,
            schema=self._database_config.source_schema,
            table=src_table,
            query=src_mismatch_sample_query,
            options=None,
        )
        tgt_data = self._target.read_data(
            catalog=self._database_config.target_catalog,
            schema=self._database_config.target_schema,
            table=tgt_table,
            query=tgt_mismatch_sample_query,
            options=None,
        )

        return capture_mismatch_data_and_columns(source=src_data, target=tgt_data, key_columns=key_columns)

    def _reconcile_threshold_data(
        self,
        table_conf: Table,
        src_schema: list[Schema],
        tgt_schema: list[Schema],
    ):

        src_data, tgt_data = self._get_threshold_data(table_conf, src_schema, tgt_schema)

        source_view = f"source_{table_conf.source_name}_df_threshold_vw"
        target_view = f"target_{table_conf.target_name}_df_threshold_vw"

        src_data.createOrReplaceTempView(source_view)
        tgt_data.createOrReplaceTempView(target_view)

        return self._compute_threshold_comparison(table_conf, src_schema)

    def _get_threshold_data(
        self,
        table_conf: Table,
        src_schema: list[Schema],
        tgt_schema: list[Schema],
    ) -> tuple[DataFrame, DataFrame]:
        src_threshold_query = ThresholdQueryBuilder(
            table_conf, src_schema, "source", self._source_engine
        ).build_threshold_query()
        tgt_threshold_query = ThresholdQueryBuilder(
            table_conf, tgt_schema, "target", self._target_engine
        ).build_threshold_query()

        src_data = self._source.read_data(
            catalog=self._database_config.source_catalog,
            schema=self._database_config.source_schema,
            table=table_conf.source_name,
            query=src_threshold_query,
            options=table_conf.jdbc_reader_options,
        )
        tgt_data = self._target.read_data(
            catalog=self._database_config.target_catalog,
            schema=self._database_config.target_schema,
            table=table_conf.target_name,
            query=tgt_threshold_query,
            options=table_conf.jdbc_reader_options,
        )

        return src_data, tgt_data

    def _compute_threshold_comparison(self, table_conf: Table, src_schema: list[Schema]) -> ThresholdOutput:
        threshold_comparison_query = ThresholdQueryBuilder(
            table_conf, src_schema, "target", self._target_engine
        ).build_comparison_query()

        threshold_result = self._target.read_data(
            catalog=self._database_config.target_catalog,
            schema=self._database_config.target_schema,
            table=table_conf.target_name,
            query=threshold_comparison_query,
            options=table_conf.jdbc_reader_options,
        )
        threshold_columns = table_conf.get_threshold_columns("source")
        failed_where_cond = " OR ".join([name + "_match = 'Failed'" for name in threshold_columns])
        mismatched_df = threshold_result.filter(failed_where_cond)
        mismatched_count = mismatched_df.count()
        threshold_df = None
        if mismatched_count > 0:
            threshold_df = mismatched_df.limit(_SAMPLE_ROWS)

        return ThresholdOutput(threshold_df=threshold_df, threshold_mismatch_count=mismatched_count)

    def get_record_count(self, table_conf: Table, report_type: str) -> ReconcileRecordCount:
        if report_type != "schema":
            source_count_query = CountQueryBuilder(table_conf, "source", self._source_engine).build_query()
            target_count_query = CountQueryBuilder(table_conf, "target", self._target_engine).build_query()
            source_count_row = self._source.read_data(
                catalog=self._database_config.source_catalog,
                schema=self._database_config.source_schema,
                table=table_conf.source_name,
                query=source_count_query,
                options=None,
            ).first()
            target_count_row = self._target.read_data(
                catalog=self._database_config.target_catalog,
                schema=self._database_config.target_schema,
                table=table_conf.target_name,
                query=target_count_query,
                options=None,
            ).first()

            source_count = int(source_count_row[0]) if source_count_row is not None else 0
            target_count = int(target_count_row[0]) if target_count_row is not None else 0

            return ReconcileRecordCount(source=int(source_count), target=int(target_count))
        return ReconcileRecordCount()


def _get_schema(
    source: DataSource,
    target: DataSource,
    table_conf: Table,
    database_config: DatabaseConfig,
) -> tuple[list[Schema], list[Schema]]:
    src_schema = source.get_schema(
        catalog=database_config.source_catalog,
        schema=database_config.source_schema,
        table=table_conf.source_name,
    )
    tgt_schema = target.get_schema(
        catalog=database_config.target_catalog,
        schema=database_config.target_schema,
        table=table_conf.target_name,
    )

    return src_schema, tgt_schema


def _run_reconcile_data(
    reconciler: Reconciliation,
    table_conf: Table,
    src_schema: list[Schema],
    tgt_schema: list[Schema],
) -> DataReconcileOutput:
    try:
        return reconciler.reconcile_data(table_conf=table_conf, src_schema=src_schema, tgt_schema=tgt_schema)
    except DataSourceRuntimeException as e:
        return DataReconcileOutput(exception=str(e))


def _run_reconcile_schema(
    reconciler: Reconciliation,
    table_conf: Table,
    src_schema: list[Schema],
    tgt_schema: list[Schema],
):
    try:
        return reconciler.reconcile_schema(table_conf=table_conf, src_schema=src_schema, tgt_schema=tgt_schema)
    except PySparkException as e:
        return SchemaReconcileOutput(is_valid=False, exception=str(e))


def _run_reconcile_aggregates(
    reconciler: Reconciliation,
    table_conf: Table,
    src_schema: list[Schema],
    tgt_schema: list[Schema],
) -> list[AggregateQueryOutput]:
    try:
        return reconciler.reconcile_aggregates(table_conf, src_schema, tgt_schema)
    except DataSourceRuntimeException as e:
        return [AggregateQueryOutput(reconcile_output=DataReconcileOutput(exception=str(e)), rule=None)]


if __name__ == "__main__":
    if "DATABRICKS_RUNTIME_VERSION" not in os.environ:
        raise SystemExit("Only intended to run in Databricks Runtime")
    main(*sys.argv)



import re
import abc
import dataclasses
import shutil
from collections.abc import Iterable
from json import loads, dump
import logging
import os
from shutil import rmtree, move
from subprocess import run, CalledProcessError
import sys
from typing import Any, cast
from urllib import request
from urllib.error import URLError, HTTPError
import webbrowser
from datetime import datetime, timezone
from pathlib import Path
import xml.etree.ElementTree as ET
from zipfile import ZipFile

from databricks.labs.blueprint.installation import Installation, JsonValue
from databricks.labs.blueprint.installation import SerdeError
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.wheels import ProductInfo
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, PermissionDenied

from databricks.labs.lakebridge.config import (
    TranspileConfig,
    ReconcileConfig,
    DatabaseConfig,
    RemorphConfigs,
    ReconcileMetadataConfig,
    LSPConfigOptionV1,
)

from databricks.labs.lakebridge.deployment.configurator import ResourceConfigurator
from databricks.labs.lakebridge.deployment.installation import WorkspaceInstallation
from databricks.labs.lakebridge.reconcile.constants import ReconReportType, ReconSourceType
from databricks.labs.lakebridge.transpiler.lsp.lsp_engine import LSPConfig

logger = logging.getLogger(__name__)

TRANSPILER_WAREHOUSE_PREFIX = "Lakebridge Transpiler Validation"


class TranspilerInstaller(abc.ABC):

    @classmethod
    def labs_path(cls) -> Path:
        return Path.home() / ".databricks" / "labs"

    @classmethod
    def transpilers_path(cls) -> Path:
        return cls.labs_path() / "remorph-transpilers"

    @classmethod
    def install_from_pypi(cls, product_name: str, pypi_name: str, artifact: Path | None = None) -> Path | None:
        installer = WheelInstaller(product_name, pypi_name, artifact)
        return installer.install()

    @classmethod
    def install_from_maven(
        cls, product_name: str, group_id: str, artifact_id: str, artifact: Path | None = None
    ) -> Path | None:
        installer = MavenInstaller(product_name, group_id, artifact_id, artifact)
        return installer.install()

    @classmethod
    def get_installed_version(cls, product_name: str, is_transpiler=True) -> str | None:
        product_path = (cls.transpilers_path() if is_transpiler else cls.labs_path()) / product_name
        current_version_path = product_path / "state" / "version.json"
        if not current_version_path.exists():
            return None
        text = current_version_path.read_text("utf-8")
        data: dict[str, Any] = loads(text)
        version: str | None = data.get("version", None)
        if not version or not version.startswith("v"):
            return None
        return version[1:]

    _version_pattern = re.compile(r"[_-](\d+(?:[.\-_]\w*\d+)+)")

    @classmethod
    def get_local_artifact_version(cls, artifact: Path) -> str | None:
        # TODO: Get the version from the metadata inside the artifact rather than relying on the filename.
        match = cls._version_pattern.search(artifact.stem)
        if not match:
            return None
        group = match.group(0)
        if not group:
            return None
        # TODO: Update the regex to take care of these trimming scenarios.
        if group.startswith('-'):
            group = group[1:]
        if group.endswith("-py3"):
            group = group[:-4]
        return group

    @classmethod
    def all_transpiler_configs(cls) -> dict[str, LSPConfig]:
        all_configs = cls._all_transpiler_configs()
        return {config.name: config for config in all_configs}

    @classmethod
    def all_transpiler_names(cls) -> set[str]:
        all_configs = cls.all_transpiler_configs()
        return set(all_configs.keys())

    @classmethod
    def all_dialects(cls) -> set[str]:
        all_dialects: set[str] = set()
        for config in cls._all_transpiler_configs():
            all_dialects = all_dialects.union(config.remorph.dialects)
        return all_dialects

    @classmethod
    def transpilers_with_dialect(cls, dialect: str) -> set[str]:
        configs = filter(lambda cfg: dialect in cfg.remorph.dialects, cls.all_transpiler_configs().values())
        return set(config.name for config in configs)

    @classmethod
    def transpiler_config_path(cls, transpiler_name) -> Path:
        config = cls.all_transpiler_configs().get(transpiler_name, None)
        if not config:
            raise ValueError(f"No such transpiler: {transpiler_name}")
        return config.path

    @classmethod
    def transpiler_config_options(cls, transpiler_name, source_dialect) -> list[LSPConfigOptionV1]:
        config = cls.all_transpiler_configs().get(transpiler_name, None)
        if not config:
            return []  # gracefully returns an empty list, since this can only happen during testing
        return config.options_for_dialect(source_dialect)

    @classmethod
    def _all_transpiler_configs(cls) -> Iterable[LSPConfig]:
        path = cls.transpilers_path()
        if path.exists():
            all_files = os.listdir(path)
            for file in all_files:
                config = cls._transpiler_config(cls.transpilers_path() / file)
                if config:
                    yield config

    @classmethod
    def _transpiler_config(cls, path: Path) -> LSPConfig | None:
        if not path.is_dir() or not (path / "lib").is_dir():
            return None
        config_path = path / "lib" / "config.yml"
        if not config_path.is_file():
            return None
        try:
            return LSPConfig.load(config_path)
        except ValueError as e:
            logger.error(f"Could not load config: {path!s}", exc_info=e)
            return None

    @classmethod
    def _store_product_state(cls, product_path: Path, version: str) -> None:
        state_path = product_path / "state"
        state_path.mkdir()
        version_data = {"version": f"v{version}", "date": datetime.now(timezone.utc).isoformat()}
        version_path = state_path / "version.json"
        with version_path.open("w", encoding="utf-8") as f:
            dump(version_data, f)
            f.write("\n")


class WheelInstaller(TranspilerInstaller):

    @classmethod
    def get_latest_artifact_version_from_pypi(cls, product_name: str) -> str | None:
        try:
            with request.urlopen(f"https://pypi.org/pypi/{product_name}/json") as server:
                text: bytes = server.read()
            data: dict[str, Any] = loads(text)
            return data.get("info", {}).get('version', None)
        except HTTPError as e:
            logger.error(f"Error while fetching PyPI metadata: {product_name}", exc_info=e)
            return None

    @classmethod
    def download_artifact_from_pypi(cls, product_name: str, version: str, target: Path, extension="whl") -> int:
        suffix = "-py3-none-any.whl" if extension == "whl" else ".tar.gz" if extension == "tar" else f".{extension}"
        filename = f"{product_name.replace('-', '_')}-{version}{suffix}"
        url = f"https://pypi.debian.net/{product_name}/{filename}"
        try:
            path, _ = request.urlretrieve(url)
            logger.info(f"Successfully downloaded {path}")
            if not target.exists():
                logger.info(f"Moving {path} to {target!s}")
                move(path, target)
            return 0
        except URLError as e:
            logger.error("While downloading from pypi", exc_info=e)
            return -1

    def __init__(self, product_name: str, pypi_name: str, artifact: Path | None = None):
        self._product_name = product_name
        self._pypi_name = pypi_name
        self._artifact = artifact

    def install(self) -> Path | None:
        return self._install_checking_versions()

    def _install_checking_versions(self) -> Path | None:
        latest_version = (
            self.get_local_artifact_version(self._artifact)
            if self._artifact
            else self.get_latest_artifact_version_from_pypi(self._pypi_name)
        )
        if latest_version is None:
            logger.warning(f"Could not determine the latest version of {self._pypi_name}")
            logger.error(f"Failed to install transpiler: {self._product_name}")
            return None
        installed_version = self.get_installed_version(self._product_name)
        if installed_version == latest_version:
            logger.info(f"{self._pypi_name} v{latest_version} already installed")
            return None
        return self._install_latest_version(latest_version)

    def _install_latest_version(self, version: str) -> Path | None:
        logger.info(f"Installing Databricks {self._product_name} transpiler v{version}")
        # use type(self) to workaround a mock bug on class methods
        self._product_path = type(self).transpilers_path() / self._product_name
        backup_path = Path(f"{self._product_path!s}-saved")
        if self._product_path.exists():
            os.rename(self._product_path, backup_path)
        self._product_path.mkdir(parents=True, exist_ok=True)
        self._install_path = self._product_path / "lib"
        self._install_path.mkdir(exist_ok=True)
        try:
            result = self._unsafe_install_latest_version(version)
            logger.info(f"Successfully installed {self._pypi_name} v{version}")
            if backup_path.exists():
                rmtree(backup_path)
            return result
        except (CalledProcessError, ValueError) as e:
            logger.error(f"Failed to install {self._pypi_name} v{version}", exc_info=e)
            rmtree(self._product_path)
            if backup_path.exists():
                os.rename(backup_path, self._product_path)
            return None

    def _unsafe_install_latest_version(self, version: str) -> Path | None:
        self._create_venv()
        self._install_with_pip()
        self._copy_lsp_resources()
        return self._post_install(version)

    def _create_venv(self) -> None:
        cwd = os.getcwd()
        try:
            os.chdir(self._install_path)
            self._unsafe_create_venv()
        finally:
            os.chdir(cwd)

    def _unsafe_create_venv(self) -> None:
        # using the venv module doesn't work (maybe it's not possible to create a venv from a venv ?)
        # so falling back to something that works
        # for some reason this requires shell=True, so pass full cmd line
        cmd_line = f"{sys.executable} -m venv .venv"
        completed = run(cmd_line, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, shell=True, check=False)
        if completed.returncode:
            logger.error(f"Failed to create venv, error code: {completed.returncode}")
            if completed.stdout:
                for line in completed.stdout:
                    logger.error(line)
            if completed.stderr:
                for line in completed.stderr:
                    logger.error(line)
        completed.check_returncode()
        self._venv = self._install_path / ".venv"
        self._site_packages = self._locate_site_packages()

    def _locate_site_packages(self) -> Path:
        # can't use sysconfig because it only works for currently running python
        if sys.platform == "win32":
            return self._locate_site_packages_windows()
        return self._locate_site_packages_linux_or_macos()

    def _locate_site_packages_windows(self) -> Path:
        packages = self._venv / "Lib" / "site-packages"
        if packages.exists():
            return packages
        raise ValueError(f"Could not locate 'site-packages' for {self._venv!s}")

    def _locate_site_packages_linux_or_macos(self) -> Path:
        lib = self._venv / "lib"
        for dir_ in os.listdir(lib):
            if dir_.startswith("python"):
                packages = lib / dir_ / "site-packages"
                if packages.exists():
                    return packages
        raise ValueError(f"Could not locate 'site-packages' for {self._venv!s}")

    def _install_with_pip(self) -> None:
        cwd = os.getcwd()
        try:
            os.chdir(self._install_path)
            # the way to call pip from python is highly sensitive to os and source type
            if self._artifact:
                self._install_local_artifact()
            else:
                self._install_remote_artifact()
        finally:
            os.chdir(cwd)

    def _install_local_artifact(self) -> None:
        pip = self._locate_pip()
        pip = pip.relative_to(self._install_path)
        target = self._site_packages
        target = target.relative_to(self._install_path)
        if sys.platform == "win32":
            command = f"{pip!s} install {self._artifact!s} -t {target!s}"
            completed = run(command, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, shell=False, check=False)
        else:
            command = f"'{pip!s}' install '{self._artifact!s}' -t '{target!s}'"
            completed = run(command, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, shell=True, check=False)
        # checking return code later makes debugging easier
        completed.check_returncode()

    def _install_remote_artifact(self) -> None:
        pip = self._locate_pip()
        pip = pip.relative_to(self._install_path)
        target = self._site_packages
        target = target.relative_to(self._install_path)
        if sys.platform == "win32":
            args = [str(pip), "install", self._pypi_name, "-t", str(target)]
            completed = run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, shell=False, check=False)
        else:
            command = f"'{pip!s}' install {self._pypi_name} -t '{target!s}'"
            completed = run(command, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, shell=True, check=False)
        # checking return code later makes debugging easier
        completed.check_returncode()

    def _locate_pip(self) -> Path:
        return self._venv / "Scripts" / "pip3.exe" if sys.platform == "win32" else self._venv / "bin" / "pip3"

    def _copy_lsp_resources(self):
        lsp = self._site_packages / "lsp"
        if not lsp.exists():
            raise ValueError("Installed transpiler is missing a 'lsp' folder")
        shutil.copytree(lsp, self._install_path, dirs_exist_ok=True)

    def _post_install(self, version: str) -> Path | None:
        config = self._install_path / "config.yml"
        if not config.exists():
            raise ValueError("Installed transpiler is missing a 'config.yml' file in its 'lsp' folder")
        install_ext = "ps1" if sys.platform == "win32" else "sh"
        install_script = f"installer.{install_ext}"
        installer = self._install_path / install_script
        if installer.exists():
            self._run_custom_installer(installer)
        self._store_product_state(product_path=self._product_path, version=version)
        return self._install_path

    def _run_custom_installer(self, installer):
        args = [str(installer)]
        run(args, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, cwd=str(self._install_path), check=True)


class MavenInstaller(TranspilerInstaller):
    # Maven Central, base URL.
    _maven_central_repo: str = "https://repo.maven.apache.org/maven2/"

    @classmethod
    def _artifact_base_url(cls, group_id: str, artifact_id: str) -> str:
        """Construct the base URL for a Maven artifact."""
        # Reference: https://maven.apache.org/repositories/layout.html
        group_path = group_id.replace(".", "/")
        return f"{cls._maven_central_repo}{group_path}/{artifact_id}/"

    @classmethod
    def artifact_metadata_url(cls, group_id: str, artifact_id: str) -> str:
        """Get the metadata URL for a Maven artifact."""
        # TODO: Unit test this method.
        return f"{cls._artifact_base_url(group_id, artifact_id)}maven-metadata.xml"

    @classmethod
    def artifact_url(
        cls, group_id: str, artifact_id: str, version: str, classifier: str | None = None, extension: str = "jar"
    ) -> str:
        """Get the URL for a versioned Maven artifact."""
        # TODO: Unit test this method, including classifier and extension.
        _classifier = f"-{classifier}" if classifier else ""
        artifact_base_url = cls._artifact_base_url(group_id, artifact_id)
        return f"{artifact_base_url}{version}/{artifact_id}-{version}{_classifier}.{extension}"

    @classmethod
    def get_current_maven_artifact_version(cls, group_id: str, artifact_id: str) -> str | None:
        url = cls.artifact_metadata_url(group_id, artifact_id)
        try:
            with request.urlopen(url) as server:
                text = server.read()
        except HTTPError as e:
            logger.error(f"Error while fetching maven metadata: {group_id}:{artifact_id}", exc_info=e)
            return None
        logger.debug(f"Maven metadata for {group_id}:{artifact_id}: {text}")
        return cls._extract_latest_release_version(text)

    @classmethod
    def _extract_latest_release_version(cls, maven_metadata: str) -> str | None:
        """Extract the latest release version from Maven metadata."""
        # Reference: https://maven.apache.org/repositories/metadata.html#The_A_Level_Metadata
        # TODO: Unit test this method, to verify the sequence of things it checks for.
        root = ET.fromstring(maven_metadata)
        for label in ("release", "latest"):
            version = root.findtext(f"./versioning/{label}")
            if version is not None:
                return version
        return root.findtext("./versioning/versions/version[last()]")

    @classmethod
    def download_artifact_from_maven(
        cls,
        group_id: str,
        artifact_id: str,
        version: str,
        target: Path,
        classifier: str | None = None,
        extension: str = "jar",
    ) -> bool:
        if target.exists():
            logger.warning(f"Skipping download of {group_id}:{artifact_id}:{version}; target already exists: {target}")
            return True
        url = cls.artifact_url(group_id, artifact_id, version, classifier, extension)
        try:
            path, _ = request.urlretrieve(url)
            logger.debug(f"Downloaded maven artefact from {url} to {path}")
        except URLError as e:
            logger.error(f"Unable to download maven artefact: {group_id}:{artifact_id}:{version}", exc_info=e)
            return False
        logger.debug(f"Moving {path} to {target}")
        move(path, target)
        logger.info(f"Successfully installed: {group_id}:{artifact_id}:{version}")
        return True

    def __init__(self, product_name: str, group_id: str, artifact_id: str, artifact: Path | None = None):
        self._product_name = product_name
        self._group_id = group_id
        self._artifact_id = artifact_id
        self._artifact = artifact

    def install(self) -> Path | None:
        return self._install_checking_versions()

    def _install_checking_versions(self) -> Path | None:
        if self._artifact:
            latest_version = self.get_local_artifact_version(self._artifact)
        else:
            latest_version = self.get_current_maven_artifact_version(self._group_id, self._artifact_id)
        if latest_version is None:
            logger.warning(f"Could not determine the latest version of Databricks {self._product_name} transpiler")
            logger.error("Failed to install transpiler: Databricks {self._product_name} transpiler")
            return None
        installed_version = self.get_installed_version(self._product_name)
        if installed_version == latest_version:
            logger.info(f"Databricks {self._product_name} transpiler v{latest_version} already installed")
            return None
        return self._install_version(latest_version)

    def _install_version(self, version: str) -> Path | None:
        logger.info(f"Installing Databricks {self._product_name} transpiler v{version}")
        # use type(self) to workaround a mock bug on class methods
        self._product_path = type(self).transpilers_path() / self._product_name
        backup_path = Path(f"{self._product_path!s}-saved")
        if backup_path.exists():
            rmtree(backup_path)
        if self._product_path.exists():
            os.rename(self._product_path, backup_path)
        self._product_path.mkdir(parents=True)
        self._install_path = self._product_path / "lib"
        self._install_path.mkdir()
        try:
            if self._unsafe_install_version(version):
                logger.info(f"Successfully installed {self._product_name} v{version}")
                self._store_product_state(self._product_path, version)
                if backup_path.exists():
                    rmtree(backup_path)
                return self._product_path
        except (KeyError, ValueError) as e:
            logger.error(f"Failed to install Databricks {self._product_name} transpiler v{version}", exc_info=e)
        rmtree(self._product_path)
        if backup_path.exists():
            os.rename(backup_path, self._product_path)
        return None

    def _unsafe_install_version(self, version: str) -> bool:
        jar_file_path = self._install_path / f"{self._artifact_id}.jar"
        if self._artifact:
            logger.debug(f"Copying '{self._artifact!s}' to '{jar_file_path!s}'")
            shutil.copyfile(self._artifact, jar_file_path)
        elif not self.download_artifact_from_maven(self._group_id, self._artifact_id, version, jar_file_path):
            logger.error(f"Failed to install Databricks {self._product_name} transpiler v{version}")
            return False
        self._copy_lsp_config(jar_file_path)
        return True

    def _copy_lsp_config(self, jar_file_path: Path) -> None:
        with ZipFile(jar_file_path) as zip_file:
            zip_file.extract("lsp/config.yml", self._install_path)
        shutil.move(self._install_path / "lsp" / "config.yml", self._install_path / "config.yml")
        os.rmdir(self._install_path / "lsp")


class WorkspaceInstaller:
    def __init__(
        self,
        ws: WorkspaceClient,
        prompts: Prompts,
        installation: Installation,
        install_state: InstallState,
        product_info: ProductInfo,
        resource_configurator: ResourceConfigurator,
        workspace_installation: WorkspaceInstallation,
        environ: dict[str, str] | None = None,
    ):
        self._ws = ws
        self._prompts = prompts
        self._installation = installation
        self._install_state = install_state
        self._product_info = product_info
        self._resource_configurator = resource_configurator
        self._ws_installation = workspace_installation

        if not environ:
            environ = dict(os.environ.items())

        if "DATABRICKS_RUNTIME_VERSION" in environ:
            msg = "WorkspaceInstaller is not supposed to be executed in Databricks Runtime"
            raise SystemExit(msg)

    def run(self, module: str, config: RemorphConfigs | None = None, artifact: str | None = None) -> RemorphConfigs:
        logger.debug(f"Initializing workspace installation for module: {module} (config: {config})")
        if module == "transpile" and artifact:
            self.install_artifact(artifact)
        elif module in {"transpile", "all"}:
            self.install_bladebridge()
            self.install_morpheus()
        if not config:
            config = self.configure(module)
        if self._is_testing():
            return config
        self._ws_installation.install(config)
        logger.info("Installation completed successfully! Please refer to the documentation for the next steps.")
        return config

    @classmethod
    def install_bladebridge(cls, artifact: Path | None = None):
        local_name = "bladebridge"
        pypi_name = "databricks-bb-plugin"
        TranspilerInstaller.install_from_pypi(local_name, pypi_name, artifact)

    @classmethod
    def install_morpheus(cls, artifact: Path | None = None):
        java_version = cls.get_java_version()
        if java_version is None or java_version < 110:
            logger.warning(
                "This software requires Java 11 or above. Please install Java and re-run 'install-transpile'."
            )
            return
        product_name = "databricks-morph-plugin"
        group_id = "com.databricks.labs"
        artifact_id = product_name
        TranspilerInstaller.install_from_maven(product_name, group_id, artifact_id, artifact)

    @classmethod
    def install_artifact(cls, artifact: str):
        path = Path(artifact)
        if not path.exists():
            logger.error(f"Could not locate artifact {artifact}")
            return
        if "databricks-morph-plugin" in path.name:
            cls.install_morpheus(path)
        elif "databricks_bb_plugin" in path.name:
            cls.install_bladebridge(path)
        else:
            logger.fatal(f"Cannot install unsupported artifact: {artifact}")

    @classmethod
    def get_java_version(cls) -> int | None:
        completed = run(["java", "-version"], shell=False, capture_output=True, check=False)
        try:
            completed.check_returncode()
        except CalledProcessError:
            return None
        result = completed.stderr.decode("utf-8")
        start = result.find(" version ")
        if start < 0:
            return None
        start = result.find('"', start + 1)
        if start < 0:
            return None
        end = result.find('"', start + 1)
        if end < 0:
            return None
        version = result[start + 1 : end]
        parts = version.split('.')
        return int(parts[0] + parts[1])

    def configure(self, module: str) -> RemorphConfigs:
        match module:
            case "transpile":
                logger.info("Configuring lakebridge `transpile`.")
                return RemorphConfigs(self._configure_transpile(), None)
            case "reconcile":
                logger.info("Configuring lakebridge `reconcile`.")
                return RemorphConfigs(None, self._configure_reconcile())
            case "all":
                logger.info("Configuring lakebridge `transpile` and `reconcile`.")
                return RemorphConfigs(
                    self._configure_transpile(),
                    self._configure_reconcile(),
                )
            case _:
                raise ValueError(f"Invalid input: {module}")

    def _is_testing(self):
        return self._product_info.product_name() != "lakebridge"

    def _configure_transpile(self) -> TranspileConfig:
        try:
            self._installation.load(TranspileConfig)
            logger.info("Lakebridge `transpile` is already installed on this workspace.")
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                raise SystemExit("Lak `transpile` is already installed and no override has been requested. Exiting...")
        except NotFound:
            logger.info("Couldn't find existing `transpile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `transpile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        config = self._configure_new_transpile_installation()
        logger.info("Finished configuring lakebridge `transpile`.")
        return config

    def _configure_new_transpile_installation(self) -> TranspileConfig:
        default_config = self._prompt_for_new_transpile_installation()
        runtime_config = None
        catalog_name = "remorph"
        schema_name = "transpiler"
        if not default_config.skip_validation:
            catalog_name = self._configure_catalog()
            schema_name = self._configure_schema(catalog_name, "transpile")
            self._has_necessary_access(catalog_name, schema_name)
            warehouse_id = self._resource_configurator.prompt_for_warehouse_setup(TRANSPILER_WAREHOUSE_PREFIX)
            runtime_config = {"warehouse_id": warehouse_id}

        config = dataclasses.replace(
            default_config,
            catalog_name=catalog_name,
            schema_name=schema_name,
            sdk_config=runtime_config,
        )
        self._save_config(config)
        return config

    def _all_installed_dialects(self) -> list[str]:
        return sorted(TranspilerInstaller.all_dialects())

    def _transpilers_with_dialect(self, dialect: str) -> list[str]:
        return sorted(TranspilerInstaller.transpilers_with_dialect(dialect))

    def _transpiler_config_path(self, transpiler: str) -> Path:
        return TranspilerInstaller.transpiler_config_path(transpiler)

    def _prompt_for_new_transpile_installation(self) -> TranspileConfig:
        install_later = "Set it later"
        # TODO tidy this up, logger might not display the below in console...
        logger.info("Please answer a few questions to configure lakebridge `transpile`")
        all_dialects = [install_later] + self._all_installed_dialects()
        source_dialect: str | None = self._prompts.choice("Select the source dialect:", all_dialects, sort=False)
        if source_dialect == install_later:
            source_dialect = None
        transpiler_name: str | None = None
        transpiler_config_path: Path | None = None
        if source_dialect:
            transpilers = self._transpilers_with_dialect(source_dialect)
            if len(transpilers) > 1:
                transpilers = [install_later] + transpilers
                transpiler_name = self._prompts.choice("Select the transpiler:", transpilers, sort=False)
                if transpiler_name == install_later:
                    transpiler_name = None
            else:
                transpiler_name = next(t for t in transpilers)
                # TODO Change name for bladebridge
                logger.info(f"lakebridge will use the {transpiler_name} transpiler")
            if transpiler_name:
                transpiler_config_path = self._transpiler_config_path(transpiler_name)
        transpiler_options: dict[str, JsonValue] | None = None
        if transpiler_config_path:
            transpiler_options = self._prompt_for_transpiler_options(
                cast(str, transpiler_name), cast(str, source_dialect)
            )
        input_source: str | None = self._prompts.question(
            "Enter input SQL path (directory/file)", default=install_later
        )
        if input_source == install_later:
            input_source = None
        output_folder = self._prompts.question("Enter output directory", default="transpiled")
        # When defaults are passed along we need to use absolute paths to avoid issues with relative paths
        if output_folder == "transpiled":
            output_folder = str(Path.cwd() / "transpiled")
        error_file_path = self._prompts.question("Enter error file path", default="errors.log")
        if error_file_path == "errors.log":
            error_file_path = str(Path.cwd() / "errors.log")

        run_validation = self._prompts.confirm(
            "Would you like to validate the syntax and semantics of the transpiled queries?"
        )

        return TranspileConfig(
            transpiler_config_path=str(transpiler_config_path) if transpiler_config_path is not None else None,
            transpiler_options=transpiler_options,
            source_dialect=source_dialect,
            skip_validation=(not run_validation),
            input_source=input_source,
            output_folder=output_folder,
            error_file_path=error_file_path,
        )

    def _prompt_for_transpiler_options(self, transpiler_name: str, source_dialect: str) -> dict[str, Any] | None:
        config_options = TranspilerInstaller.transpiler_config_options(transpiler_name, source_dialect)
        if len(config_options) == 0:
            return None
        return {option.flag: option.prompt_for_value(self._prompts) for option in config_options}

    def _configure_catalog(
        self,
    ) -> str:
        return self._resource_configurator.prompt_for_catalog_setup()

    def _configure_schema(
        self,
        catalog: str,
        default_schema_name: str,
    ) -> str:
        return self._resource_configurator.prompt_for_schema_setup(
            catalog,
            default_schema_name,
        )

    def _configure_reconcile(self) -> ReconcileConfig:
        try:
            self._installation.load(ReconcileConfig)
            logger.info("lakebridge `reconcile` is already installed on this workspace.")
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                raise SystemExit(
                    "lakebridge `reconcile` is already installed and no override has been requested. Exiting..."
                )
        except NotFound:
            logger.info("Couldn't find existing `reconcile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `reconcile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        config = self._configure_new_reconcile_installation()
        logger.info("Finished configuring lakebridge `reconcile`.")
        return config

    def _configure_new_reconcile_installation(self) -> ReconcileConfig:
        default_config = self._prompt_for_new_reconcile_installation()
        self._save_config(default_config)
        return default_config

    def _prompt_for_new_reconcile_installation(self) -> ReconcileConfig:
        logger.info("Please answer a few questions to configure lakebridge `reconcile`")
        data_source = self._prompts.choice(
            "Select the Data Source:", [source_type.value for source_type in ReconSourceType]
        )
        report_type = self._prompts.choice(
            "Select the report type:", [report_type.value for report_type in ReconReportType]
        )
        scope_name = self._prompts.question(
            f"Enter Secret scope name to store `{data_source.capitalize()}` connection details / secrets",
            default=f"remorph_{data_source}",
        )

        db_config = self._prompt_for_reconcile_database_config(data_source)
        metadata_config = self._prompt_for_reconcile_metadata_config()

        return ReconcileConfig(
            data_source=data_source,
            report_type=report_type,
            secret_scope=scope_name,
            database_config=db_config,
            metadata_config=metadata_config,
        )

    def _prompt_for_reconcile_database_config(self, source) -> DatabaseConfig:
        source_catalog = None
        if source == ReconSourceType.SNOWFLAKE.value:
            source_catalog = self._prompts.question(f"Enter source catalog name for `{source.capitalize()}`")

        schema_prompt = f"Enter source schema name for `{source.capitalize()}`"
        if source in {ReconSourceType.ORACLE.value}:
            schema_prompt = f"Enter source database name for `{source.capitalize()}`"

        source_schema = self._prompts.question(schema_prompt)
        target_catalog = self._prompts.question("Enter target catalog name for Databricks")
        target_schema = self._prompts.question("Enter target schema name for Databricks")

        return DatabaseConfig(
            source_schema=source_schema,
            target_catalog=target_catalog,
            target_schema=target_schema,
            source_catalog=source_catalog,
        )

    def _prompt_for_reconcile_metadata_config(self) -> ReconcileMetadataConfig:
        logger.info("Configuring reconcile metadata.")
        catalog = self._configure_catalog()
        schema = self._configure_schema(
            catalog,
            "reconcile",
        )
        volume = self._configure_volume(catalog, schema, "reconcile_volume")
        self._has_necessary_access(catalog, schema, volume)
        return ReconcileMetadataConfig(catalog=catalog, schema=schema, volume=volume)

    def _configure_volume(
        self,
        catalog: str,
        schema: str,
        default_volume_name: str,
    ) -> str:
        return self._resource_configurator.prompt_for_volume_setup(
            catalog,
            schema,
            default_volume_name,
        )

    def _save_config(self, config: TranspileConfig | ReconcileConfig):
        logger.info(f"Saving configuration file {config.__file__}")
        self._installation.save(config)
        ws_file_url = self._installation.workspace_link(config.__file__)
        if self._prompts.confirm(f"Open config file {ws_file_url} in the browser?"):
            webbrowser.open(ws_file_url)

    def _has_necessary_access(self, catalog_name: str, schema_name: str, volume_name: str | None = None):
        self._resource_configurator.has_necessary_access(catalog_name, schema_name, volume_name)



from __future__ import annotations
import abc
from pathlib import Path

from databricks.labs.lakebridge.config import TranspileResult, TranspileConfig


class TranspileEngine(abc.ABC):

    @classmethod
    def load_engine(cls, transpiler_config_path: Path) -> TranspileEngine:
        # TODO remove this once sqlglot transpiler is pluggable
        if str(transpiler_config_path) == "sqlglot":
            # pylint: disable=import-outside-toplevel, cyclic-import
            from databricks.labs.lakebridge.transpiler.sqlglot.sqlglot_engine import SqlglotEngine

            return SqlglotEngine()
        if not transpiler_config_path.exists():
            raise ValueError(
                f"Error: Invalid value for '--transpiler-config-path': '{str(transpiler_config_path)}', file does not exist."
            )
        # pylint: disable=import-outside-toplevel, cyclic-import
        from databricks.labs.lakebridge.transpiler.lsp.lsp_engine import LSPEngine

        return LSPEngine.from_config_path(transpiler_config_path)

    @abc.abstractmethod
    async def initialize(self, config: TranspileConfig) -> None: ...

    @abc.abstractmethod
    async def shutdown(self) -> None: ...

    @abc.abstractmethod
    async def transpile(
        self, source_dialect: str, target_dialect: str, source_code: str, file_path: Path
    ) -> TranspileResult: ...

    @property
    @abc.abstractmethod
    def supported_dialects(self) -> list[str]: ...

    def check_source_dialect(self, source_dialect: str | None) -> None:
        if source_dialect not in self.supported_dialects:
            raise ValueError(
                f"Invalid value for '--source-dialect': '{source_dialect}' is not one of {self.supported_dialects}."
            )

    @abc.abstractmethod
    def is_supported_file(self, file: Path) -> bool: ...



from pathlib import Path
from collections.abc import Generator


def is_sql_file(file: str | Path) -> bool:
    """
    Checks if the given file is a SQL file.

    :param file: The name of the file to check.
    :return: True if the file is a SQL file (i.e., its extension is either .sql or .ddl), False otherwise.
    """
    file_extension = Path(file).suffix
    return file_extension.lower() in {".sql", ".ddl"}


def is_dbt_project_file(file: Path):
    # it's ok to hardcode the file name here, see https://docs.getdbt.com/reference/dbt_project.yml
    return file.name == "dbt_project.yml"


def make_dir(path: str | Path) -> None:
    """
    Creates a directory at the specified path if it does not already exist.

    :param path: The path where the directory should be created.
    """
    Path(path).mkdir(parents=True, exist_ok=True)


def dir_walk(root: Path):
    """
    Walks the directory tree rooted at the given path, yielding a tuple containing the root directory, a list of
    :param root: Path
    :return: tuple of  root, subdirectory , files
    """
    sub_dirs = [d for d in root.iterdir() if d.is_dir()]
    files = [f for f in root.iterdir() if f.is_file()]
    yield root, sub_dirs, files

    for each_dir in sub_dirs:
        yield from dir_walk(each_dir)


def get_sql_file(input_path: str | Path) -> Generator[Path, None, None]:
    """
    Returns Generator that yields the names of all SQL files in the given directory.
    :param input_path: Path
    :return: List of SQL files
    """
    for _, _, files in dir_walk(Path(input_path)):
        for filename in files:
            if is_sql_file(filename):
                yield filename


def read_file(filename: str | Path) -> str:
    """
    Reads the contents of the given file and returns it as a string.
    :param filename: Input File Path
    :return: File Contents as String
    """
    # pylint: disable=unspecified-encoding
    with Path(filename).open() as file:
        return file.read()



import asyncio
import dataclasses
import json
import os
import time
from pathlib import Path
from typing import NoReturn, cast

from databricks.sdk.core import with_user_agent_extra
from databricks.sdk.service.sql import CreateWarehouseRequestWarehouseType
from databricks.sdk import WorkspaceClient

from databricks.labs.blueprint.cli import App
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.blueprint.installation import JsonValue
from databricks.labs.blueprint.tui import Prompts

from databricks.labs.bladespector.analyzer import Analyzer


from databricks.labs.lakebridge.assessments.configure_assessment import (
    create_assessment_configurator,
    PROFILER_SOURCE_SYSTEM,
)

from databricks.labs.lakebridge.__about__ import __version__
from databricks.labs.lakebridge.config import TranspileConfig, LSPConfigOptionV1
from databricks.labs.lakebridge.contexts.application import ApplicationContext
from databricks.labs.lakebridge.helpers.recon_config_utils import ReconConfigPrompts
from databricks.labs.lakebridge.helpers.telemetry_utils import make_alphanum_or_semver
from databricks.labs.lakebridge.install import WorkspaceInstaller
from databricks.labs.lakebridge.install import TranspilerInstaller
from databricks.labs.lakebridge.reconcile.runner import ReconcileRunner
from databricks.labs.lakebridge.lineage import lineage_generator
from databricks.labs.lakebridge.reconcile.recon_config import RECONCILE_OPERATION_NAME, AGG_RECONCILE_OPERATION_NAME
from databricks.labs.lakebridge.transpiler.execute import transpile as do_transpile


from databricks.labs.lakebridge.transpiler.lsp.lsp_engine import LSPConfig
from databricks.labs.lakebridge.transpiler.sqlglot.sqlglot_engine import SqlglotEngine
from databricks.labs.lakebridge.transpiler.transpile_engine import TranspileEngine


lakebridge = App(__file__)
logger = get_logger(__file__)


def raise_validation_exception(msg: str) -> NoReturn:
    raise ValueError(msg)


def _installer(ws: WorkspaceClient) -> WorkspaceInstaller:
    app_context = ApplicationContext(_verify_workspace_client(ws))
    return WorkspaceInstaller(
        app_context.workspace_client,
        app_context.prompts,
        app_context.installation,
        app_context.install_state,
        app_context.product_info,
        app_context.resource_configurator,
        app_context.workspace_installation,
    )


def _create_warehouse(ws: WorkspaceClient) -> str:

    dbsql = ws.warehouses.create_and_wait(
        name=f"lakebridge-warehouse-{time.time_ns()}",
        warehouse_type=CreateWarehouseRequestWarehouseType.PRO,
        cluster_size="Small",  # Adjust size as needed
        auto_stop_mins=30,  # Auto-stop after 30 minutes of inactivity
        enable_serverless_compute=True,
        max_num_clusters=1,
    )

    if dbsql.id is None:
        raise RuntimeError(f"Failed to create warehouse {dbsql.name}")

    logger.info(f"Created warehouse with id: {dbsql.id}")
    return dbsql.id


def _remove_warehouse(ws: WorkspaceClient, warehouse_id: str):
    ws.warehouses.delete(warehouse_id)
    logger.info(f"Removed warehouse post installation with id: {warehouse_id}")


def _verify_workspace_client(ws: WorkspaceClient) -> WorkspaceClient:
    """
    [Private] Verifies and updates the workspace client configuration.
    """

    # Using reflection to set right value for _product_info for telemetry
    product_info = getattr(ws.config, '_product_info')
    if product_info[0] != "lakebridge":
        setattr(ws.config, '_product_info', ('lakebridge', __version__))

    return ws


@lakebridge.command
def transpile(
    w: WorkspaceClient,
    transpiler_config_path: str | None = None,
    source_dialect: str | None = None,
    input_source: str | None = None,
    output_folder: str | None = None,
    error_file_path: str | None = None,
    skip_validation: str | None = None,
    catalog_name: str | None = None,
    schema_name: str | None = None,
):
    """Transpiles source dialect to databricks dialect"""
    ctx = ApplicationContext(w)
    logger.debug(f"Application transpiler config: {ctx.transpile_config}")
    checker = _TranspileConfigChecker(ctx.transpile_config, ctx.prompts)
    checker.check_input_source(input_source)
    checker.check_source_dialect(source_dialect)
    checker.check_transpiler_config_path(transpiler_config_path)
    checker.check_transpiler_config_options()
    checker.check_output_folder(output_folder)
    checker.check_error_file_path(error_file_path)
    checker.check_skip_validation(skip_validation)
    checker.check_catalog_name(catalog_name)
    checker.check_schema_name(schema_name)
    config, engine = checker.check()
    result = asyncio.run(_transpile(ctx, config, engine))
    # DO NOT Modify this print statement, it is used by the CLI to display results in GO Table Template
    print(json.dumps(result))


class _TranspileConfigChecker:

    def __init__(self, config: TranspileConfig | None, prompts: Prompts):
        if not config:
            raise SystemExit("Installed transpile config not found. Please install lakebridge transpile first.")
        self._config: TranspileConfig = config
        self._prompts = prompts

    def check_input_source(self, input_source: str | None):
        if input_source == "None":
            input_source = None
        if not input_source:
            input_source = self._config.input_source
        if not input_source:
            input_source = self._prompts.question("Enter input SQL path (directory/file)")
            input_source = input_source.strip()
        if not input_source:
            raise_validation_exception("Missing '--input-source'")
        if not os.path.exists(input_source):
            raise_validation_exception(f"Invalid value for '--input-source': Path '{input_source}' does not exist.")
        logger.debug(f"Setting input_source to '{input_source}'")
        self._config = dataclasses.replace(self._config, input_source=input_source)

    def check_source_dialect(self, source_dialect: str | None):
        if source_dialect == "None":
            source_dialect = None
        if not source_dialect:
            source_dialect = self._config.source_dialect
        all_dialects = sorted(TranspilerInstaller.all_dialects())
        if source_dialect and source_dialect not in all_dialects:
            logger.error(f"'{source_dialect}' is not a supported dialect. Selecting a supported one...")
            source_dialect = None
        if not source_dialect:
            source_dialect = self._prompts.choice("Select the source dialect:", all_dialects)
        if not source_dialect:
            raise_validation_exception("Missing '--source-dialect'")
        logger.debug(f"Setting source_dialect to '{source_dialect}'")
        self._config = dataclasses.replace(self._config, source_dialect=source_dialect)

    def check_transpiler_config_path(self, transpiler_config_path: str | None):
        if transpiler_config_path == "None":
            transpiler_config_path = None
        if not transpiler_config_path:
            transpiler_config_path = self._config.transpiler_config_path
        # we allow pointing to a loose transpiler config (i.e. not installed under .databricks)
        if transpiler_config_path:
            if not os.path.exists(transpiler_config_path):
                logger.error(f"The transpiler configuration does not exist '{transpiler_config_path}'.")
                transpiler_config_path = None
        if transpiler_config_path:
            config = LSPConfig.load(Path(transpiler_config_path))
            if self._config.source_dialect not in config.remorph.dialects:
                logger.error(f"The configured transpiler does not support dialect '{self._config.source_dialect}'.")
                transpiler_config_path = None
        if not transpiler_config_path:
            transpiler_names = TranspilerInstaller.transpilers_with_dialect(cast(str, self._config.source_dialect))
            if len(transpiler_names) > 1:
                transpiler_name = self._prompts.choice("Select the transpiler:", list(transpiler_names))
            else:
                transpiler_name = next(name for name in transpiler_names)
                logger.info(f"lakebridge will use the {transpiler_name} transpiler")
            transpiler_config_path = str(TranspilerInstaller.transpiler_config_path(transpiler_name))
        logger.debug(f"Setting transpiler_config_path to '{transpiler_config_path}'")
        self._config = dataclasses.replace(self._config, transpiler_config_path=cast(str, transpiler_config_path))

    def check_transpiler_config_options(self):
        lsp_config = LSPConfig.load(Path(self._config.transpiler_config_path))
        options_to_configure = lsp_config.options_for_dialect(self._config.source_dialect) or []
        transpiler_options = self._config.transpiler_options or {}
        if len(options_to_configure) == 0:
            transpiler_options = None
        else:
            # TODO delete stale options ?
            for option in options_to_configure:
                self._check_transpiler_config_option(option, transpiler_options)
        logger.debug(f"Setting transpiler_options to {transpiler_options}")
        self._config = dataclasses.replace(self._config, transpiler_options=transpiler_options)

    def _check_transpiler_config_option(self, option: LSPConfigOptionV1, values: dict[str, JsonValue]):
        if option.flag in values.keys():
            return
        values[option.flag] = option.prompt_for_value(self._prompts)

    def check_output_folder(self, output_folder: str | None):
        output_folder = output_folder if output_folder else self._config.output_folder
        if not output_folder:
            raise_validation_exception("Missing '--output-folder'")
        if not os.path.exists(output_folder):
            os.makedirs(output_folder, exist_ok=True)
        logger.debug(f"Setting output_folder to '{output_folder}'")
        self._config = dataclasses.replace(self._config, output_folder=output_folder)

    def check_error_file_path(self, error_file_path: str | None):
        error_file_path = error_file_path if error_file_path else self._config.error_file_path
        if not error_file_path or error_file_path == "None":
            raise_validation_exception("Missing '--error-file-path'")
        if error_file_path == "errors.log":
            error_file_path = str(Path.cwd() / "errors.log")
        if not os.path.exists(Path(error_file_path).parent):
            os.makedirs(Path(error_file_path).parent, exist_ok=True)

        logger.debug(f"Setting error_file_path to '{error_file_path}'")
        self._config = dataclasses.replace(self._config, error_file_path=error_file_path)

    def check_skip_validation(self, skip_validation_str: str | None):
        skip_validation: bool | None = None
        if skip_validation_str == "None":
            skip_validation_str = None
        if skip_validation_str is not None:
            if skip_validation_str.lower() not in {"true", "false"}:
                raise_validation_exception(
                    f"Invalid value for '--skip-validation': '{skip_validation_str}' is not one of 'true', 'false'."
                )
            skip_validation = skip_validation_str.lower() == "true"
        if skip_validation is None:
            skip_validation = self._config.skip_validation
        if skip_validation is None:
            skip_validation = self._prompts.confirm(
                "Would you like to validate the syntax and semantics of the transpiled queries?"
            )
        logger.debug(f"Setting skip_validation to '{skip_validation}'")
        self._config = dataclasses.replace(self._config, skip_validation=skip_validation)

    def check_catalog_name(self, catalog_name: str | None):
        if self._config.skip_validation:
            return
        if catalog_name == "None":
            catalog_name = None
        if not catalog_name:
            catalog_name = self._config.catalog_name
        if not catalog_name:
            raise_validation_exception(
                "Missing '--catalog-name', please run 'databricks labs lakebridge install-transpile' to configure one"
            )
        logger.debug(f"Setting catalog_name to '{catalog_name}'")
        self._config = dataclasses.replace(self._config, catalog_name=catalog_name)

    def check_schema_name(self, schema_name: str | None):
        if self._config.skip_validation:
            return
        if schema_name == "None":
            schema_name = None
        if not schema_name:
            schema_name = self._config.schema_name
        if not schema_name:
            raise_validation_exception(
                "Missing '--schema-name', please run 'databricks labs lakebridge install-transpile' to configure one"
            )
        logger.debug(f"Setting schema_name to '{schema_name}'")
        self._config = dataclasses.replace(self._config, schema_name=schema_name)

    def check(self) -> tuple[TranspileConfig, TranspileEngine]:
        logger.debug(f"Checking config: {self!s}")
        # not using os.path.exists because it sometimes fails mysteriously...
        transpiler_path = self._config.transpiler_path
        if not transpiler_path or not transpiler_path.exists():
            raise_validation_exception(
                f"Invalid value for '--transpiler-config-path': Path '{self._config.transpiler_config_path}' does not exist."
            )
        engine = TranspileEngine.load_engine(transpiler_path)
        engine.check_source_dialect(self._config.source_dialect)
        if not self._config.input_source or not os.path.exists(self._config.input_source):
            raise_validation_exception(
                f"Invalid value for '--input-source': Path '{self._config.input_source}' does not exist."
            )
        # 'transpiled' will be used as output_folder if not specified
        # 'errors.log' will be used as errors file if not specified
        return self._config, engine


async def _transpile(ctx: ApplicationContext, config: TranspileConfig, engine: TranspileEngine):
    """Transpiles source dialect to databricks dialect"""
    with_user_agent_extra("cmd", "execute-transpile")
    user = ctx.current_user
    logger.debug(f"User: {user}")
    _override_workspace_client_config(ctx, config.sdk_config)
    status, errors = await do_transpile(ctx.workspace_client, engine, config)
    for error in errors:
        logger.error(f"Error Transpiling: {str(error)}")

    # Table Template in labs.yml requires the status to be list of dicts Do not change this
    logger.info(f"lakebridge Transpiler encountered {len(status)} from given {config.input_source} files.")
    return [status]


def _override_workspace_client_config(ctx: ApplicationContext, overrides: dict[str, str] | None):
    """
    Override the Workspace client's SDK config with the user provided SDK config.
    Users can provide the cluster_id and warehouse_id during the installation.
    This will update the default config object in-place.
    """
    if not overrides:
        return

    warehouse_id = overrides.get("warehouse_id")
    if warehouse_id:
        ctx.connect_config.warehouse_id = warehouse_id

    cluster_id = overrides.get("cluster_id")
    if cluster_id:
        ctx.connect_config.cluster_id = cluster_id


@lakebridge.command
def reconcile(w: WorkspaceClient):
    """[EXPERIMENTAL] Reconciles source to Databricks datasets"""
    with_user_agent_extra("cmd", "execute-reconcile")
    ctx = ApplicationContext(w)
    user = ctx.current_user
    logger.debug(f"User: {user}")
    recon_runner = ReconcileRunner(
        ctx.workspace_client,
        ctx.installation,
        ctx.install_state,
        ctx.prompts,
    )
    recon_runner.run(operation_name=RECONCILE_OPERATION_NAME)


@lakebridge.command
def aggregates_reconcile(w: WorkspaceClient):
    """[EXPERIMENTAL] Reconciles Aggregated source to Databricks datasets"""
    with_user_agent_extra("cmd", "execute-aggregates-reconcile")
    ctx = ApplicationContext(w)
    user = ctx.current_user
    logger.debug(f"User: {user}")
    recon_runner = ReconcileRunner(
        ctx.workspace_client,
        ctx.installation,
        ctx.install_state,
        ctx.prompts,
    )

    recon_runner.run(operation_name=AGG_RECONCILE_OPERATION_NAME)


@lakebridge.command
def generate_lineage(w: WorkspaceClient, source_dialect: str, input_source: str, output_folder: str):
    """[Experimental] Generates a lineage of source SQL files or folder"""
    ctx = ApplicationContext(w)
    logger.debug(f"User: {ctx.current_user}")
    engine = SqlglotEngine()
    engine.check_source_dialect(source_dialect)
    if not input_source or not os.path.exists(input_source):
        raise_validation_exception(f"Invalid value for '--input-source': Path '{input_source}' does not exist.")
    if not os.path.exists(output_folder) or output_folder in {None, ""}:
        raise_validation_exception(f"Invalid value for '--output-folder': Path '{output_folder}' does not exist.")

    lineage_generator(engine, source_dialect, input_source, output_folder)


@lakebridge.command
def configure_secrets(w: WorkspaceClient):
    """Setup reconciliation connection profile details as Secrets on Databricks Workspace"""
    recon_conf = ReconConfigPrompts(w)

    # Prompt for source
    source = recon_conf.prompt_source()

    logger.info(f"Setting up Scope, Secrets for `{source}` reconciliation")
    recon_conf.prompt_and_save_connection_details()


@lakebridge.command(is_unauthenticated=True)
def configure_database_profiler():
    """[Experimental] Install the lakebridge Assessment package"""
    prompts = Prompts()

    # Prompt for source system
    source_system = str(
        prompts.choice("Please select the source system you want to configure", PROFILER_SOURCE_SYSTEM)
    ).lower()

    # Create appropriate assessment configurator
    assessment = create_assessment_configurator(source_system=source_system, product_name="lakebridge", prompts=prompts)
    assessment.run()


@lakebridge.command()
def install_transpile(w: WorkspaceClient, artifact: str | None = None):
    """Install the lakebridge Transpilers"""
    with_user_agent_extra("cmd", "install-transpile")
    user = w.current_user
    logger.debug(f"User: {user}")
    installer = _installer(w)
    installer.run(module="transpile", artifact=artifact)


@lakebridge.command(is_unauthenticated=False)
def configure_reconcile(w: WorkspaceClient):
    """Configure the lakebridge Reconcile Package"""
    with_user_agent_extra("cmd", "configure-reconcile")
    user = w.current_user
    logger.debug(f"User: {user}")
    dbsql_id = _create_warehouse(w)
    w.config.warehouse_id = dbsql_id
    installer = _installer(w)
    installer.run(module="reconcile")
    _remove_warehouse(w, dbsql_id)


@lakebridge.command()
def analyze(w: WorkspaceClient, source_directory: str, report_file: str):
    """Run the Analyzer"""
    with_user_agent_extra("cmd", "analyze")
    ctx = ApplicationContext(w)
    prompts = ctx.prompts
    output_file = report_file
    input_folder = source_directory
    source_tech = prompts.choice("Select the source technology", Analyzer.supported_source_technologies())
    with_user_agent_extra("analyzer_source_tech", make_alphanum_or_semver(source_tech))
    user = ctx.current_user
    logger.debug(f"User: {user}")
    Analyzer.analyze(Path(input_folder), Path(output_file), source_tech)


if __name__ == "__main__":
    lakebridge()
